require 'active_model'

module GreekFire
  class Measure
    include ActiveModel::Conversion

    attr_reader :name, :description
    @stat_refreshment = []

    def self.register(name, description:nil, labels:nil, &block)
      GreekFire::Metric.register(self.new(name, description: description, labels: labels, &block))
    end

    def self.configure(&block)
      block.call(config)
    end

    def initialize(name, description:nil, labels:nil, &block)
      labels ||= {}
      @name = name
      @extra_labels = if GreekFire::Measure.config.default_labels
                        {
                            app_name: ::Rails.application.class.parent_name,
                            app_env: Rails.env.to_s
                        }
                      else
                        { }
                      end

      @labels = labels

      @description = description
      @callable = block
    end

    def labels
      return label_products(convert_label_values_to_array(@labels.dup.clear.update(@extra_labels).merge(@labels))) if @labels.is_a?(Hash)

      @labels.map.with_index { |label, i| @labels[i].dup.clear.update(@extra_labels).merge(label) }
    end

    def value(labels)
      @callable.call(labels).tap do |value|
        if !value.nil? && !value.is_a?(Numeric)
          raise ArgumentError, "Gauge value must be a number: #{value.inspect}"
        end
      end
    end

    def metrics
      labels.map do |label_set|
        val = value(label_set)
        Metric.new(name, label_set, val) unless val.nil?
      end.compact
    end

    private

    def self.config
      @config ||= OpenStruct.new(default_labels: true)
    end

    def convert_label_values_to_array(labels)
      labels.inject({}) do |doc, label_name|
        label_values = label_name[1]
        label_values = label_values.call if label_values.respond_to? :call
        label_values = [label_values] unless label_values.is_a?(Enumerable)
        doc[label_name[0]] = label_values
        doc
      end
    end

    def label_products(hsh)
      #http://stackoverflow.com/questions/9786264/all-possible-combinations-from-a-hash-of-arrays-in-ruby
      #thanks, internet
      attrs   = hsh.values
      keys    = hsh.keys

      return [{}] unless attrs.size > 0

      product = attrs[0].product(*attrs[1..-1])
      product.map{ |p| HashWithIndifferentAccess[keys.zip p] }
    end

  end
end
